#include "amd3dy.h"

void cli(void);
#pragma aux cli = "cli";

void sti(void);
#pragma aux sti = "sti" modify [edi];

void flushpipeline(void);
#pragma aux flushpipeline = 0x0F 0xA2 modify [eax ebx ecx edx];

static int TSC[4];

void rdtsc(int * tsc);
#pragma aux rdtsc = ".586p" \
"rdtsc" \
"mov [edi],eax" \
"mov [edi+4],edx" \
modify [eax edx] parm [edi];

double tscdiff(int * t0, int * t1);
#pragma aux tscdiff =     \
"fild qword ptr [edi]   " \
"fild qword ptr [esi]   " \
"fsubrp st(1),st	" \
value [8087] parm [edi] [esi];

#define N 512

float V[4];
float A[4][4];

void vectormul(void) {
int i,j;
double t0;
float Vout[4];

	cli();
	flushpipeline();
	rdtsc(&TSC[0]);

	for(j=0;j<N;j++)
	for(i=0;i<4;i++) {
		Vout[i] = A[i][0] * V[0]
				+ A[i][1] * V[1]
				+ A[i][2] * V[2]
				+ A[i][3] * V[3];
	}

	rdtsc(&TSC[2]);
	sti();

	t0 = tscdiff(&TSC[0],&TSC[2]);

	printf("%g clocks\n",(t0)/N);

}

float vectormul3DNow(void) {
int i,j;
double t0;
float Vout[4];

	/* Warm L1 cache */
	for(i=0;i<4;i++) {
		Vout[i] = i;
		A[i][0] = i+i;
		A[i][1] = i*i*i/32;
		A[i][2] = i+3;
		A[i][3] = i*i*(1/8);
		V[i]    = i+2;
	}

	cli();
	flushpipeline();
	rdtsc(&TSC[0]);

		_asm {
			lea eax, A
			lea ebx, V
			lea edx, Vout
			mov	ecx, 256

			nop		; alignment.

			femms

			movq	mm0, [ebx]			; V1 				V0
			movq	mm1, [ebx+8]		; V3 				V2
			movq	mm2, mm0			; V1 				V0
			pfmul	(mm0, eax)			; A01*V1 			A00*V0
			movq	mm3, mm1			; V3 				V2
			pfmulm	(mm1, eax, 8)		; A03*V3 			A02*V2

			L1:
			pfmulm	(mm2, eax, 16)		; A11*V1 			A10*V0
			pfmulm	(mm3, eax, 24)		; A13*V3 			A12*V2
			movq	mm4, [ebx]			; V1 				V0
			movq	mm5, [ebx+8]		; V3 				V2
			movq	mm6, mm4			; V1 				V0
			movq	mm7, mm5        	; V3 				V2
			pfadd	(mm0, mm1)			; A01*V1+A03*V3		A00*V0+A02*V2
			pfmulm	(mm4, eax, 32)		; A21*V1 			A20*V0
			pfadd	(mm2, mm3)			; A11*V1+A13*V3		A10*V0+A12*V2
			pfmulm	(mm5, eax, 40)		; A23*V3 			A22*V2
			pfmulm	(mm6, eax, 48)		; A31*V1 			A30*V0
			pfacc	(mm0, mm2)			; A11*V1+A13*V3+A10*V0+A12*V2
										; A01*V1+A03*V3+A00*V0+A02*V2
			movq	[edx], mm0			; store
			pfmulm	(mm7, eax, 56)		; A33*V3 			A32*V2
			pfadd	(mm4, mm5)			; A21*V1+A23*V3		A20*V0+A22*V2
			movq	mm0, [ebx]			; V1 				V0
			movq	mm1, [ebx+8]        ; V3 				V2
			movq	mm2, mm0			; V1 				V0
			pfmul	(mm0, eax)			; A01*V1 			A00*V0
			pfadd	(mm6, mm7)          ; A31*V1+A33*V3		A30*V0+A32*V2
			movq	mm3, mm1			; V3 				V2
			pfmulm	(mm1, eax, 8)		; A03*V3 			A02*V2
			pfacc	(mm4, mm6)			; A31*V1+A33*V3+A30*V0+A32*V2
										; A21*V1+A23*V3+A20*V0+A22*V2
			pfmulm	(mm2, eax, 16)		; A11*V1 			A10*V0
			pfmulm	(mm3, eax, 24)		; A13*V3 			A12*V2
			movq	[edx+8], mm4		; store
			movq	mm4, [ebx]			; V1 				V0
			movq	mm5, [ebx+8]		; V3 				V2
			movq	mm6, mm4			; V1 				V0
			movq	mm7, mm5        	; V3 				V2
			pfadd	(mm0, mm1)			; A01*V1+A03*V3		A00*V0+A02*V2
			pfmulm	(mm4, eax, 32)		; A21*V1 			A20*V0
			pfadd	(mm2, mm3)			; A11*V1+A13*V3		A10*V0+A12*V2
			pfmulm	(mm5, eax, 40)		; A23*V3 			A22*V2
			pfmulm	(mm6, eax, 48)		; A31*V1 			A30*V0
			pfacc	(mm0, mm2)			; A11*V1+A13*V3+A10*V0+A12*V2
										; A01*V1+A03*V3+A00*V0+A02*V2
			movq	[edx], mm0			; store
			pfmulm	(mm7, eax, 56)		; A33*V3 			A32*V2
			movq	mm0, [ebx]			; V1 				V0
			pfadd	(mm4, mm5)			; A21*V1+A23*V3		A20*V0+A22*V2
			movq	mm1, [ebx+8]        ; V3 				V2
			movq	mm2, mm0			; V1 				V0
			pfmul	(mm0, eax)			; A01*V1 			A00*V0
			pfadd	(mm6, mm7)          ; A31*V1+A33*V3		A30*V0+A32*V2
			movq	mm3, mm1			; V3 				V2
			pfmulm	(mm1, eax, 8)		; A03*V3 			A02*V2
			pfacc	(mm4, mm6)			; A31*V1+A33*V3+A30*V0+A32*V2
										; A21*V1+A23*V3+A20*V0+A22*V2
			movq	[edx+8], mm4		; store
			dec		ecx
			jnz		L1
			nop

			femms
		}

	rdtsc(&TSC[2]);
	sti();

	t0 = tscdiff(&TSC[0],&TSC[2]);

	return (t0/N);

}

main() {
int i;
float m,n;

//	vectormul();

	m = 100;
	for(i=0;i<16384;i++) {
		n = vectormul3DNow();
		if( n < m ) {
			m = n;
		}
	}

	printf("%13.11G clocks\n",m);

}

